/*
 * Copyright (C) 2011 Canonical, Ltd.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <neil.patel@canonical.com>
 *
 */

using GLib;

namespace Unity {

/*
 * Previews
 */

public abstract class Preview : Object, Dee.Serializable
{
  public string title
  {
    get { return _raw.title; }
    set { _raw.title = value; }
  }
  public string subtitle
  {
    get { return _raw.subtitle; }
    set { _raw.subtitle = value; }
  }
  public string description_markup
  {
    get { return _raw.description; }
    set { _raw.description = value; }
  }
  public string image_source_uri
  {
    get { return _raw.image_source_uri; }
    set { _raw.image_source_uri = value; }
  }
  public Icon? image
  {
    get { return _raw.image; }
    set { _raw.image = value; }
  }

  private Protocol.Preview? _raw;
  internal unowned Protocol.Preview? get_raw ()
  {
    return _raw;
  }

  construct
  {
    _raw = create_raw () as Protocol.Preview;
    warn_if_fail (_raw != null);
  }

  // as a virtual method this will get into our main .h file and we don't want
  // to add dep on the unity-protocol library, therefore it returns Object
  // instead of Protocol.Preview
  internal abstract Object create_raw ();

  private GenericArray<PreviewAction> _actions =
    new GenericArray<PreviewAction> ();

  public void add_action (PreviewAction action)
  {
    _actions.add (action);

    _raw.add_action (action.id, action.display_name, action.icon_hint,
                     action.layout_hint);
  }

  public void add_info (InfoHint info_hint)
  {
    info_hint.ref_sink ();
    _raw.add_info_hint (info_hint.id, info_hint.display_name,
                        info_hint.icon_hint, info_hint.data);
  }

  internal virtual HashTable<string, Variant> update_property (HashTable<string, Variant> values)
  {
    _raw.update_property (values);
    return new HashTable<string, Variant> (str_hash, str_equal);
  }

  internal unowned GenericArray<PreviewAction> get_actions ()
  {
    return _actions;
  }

  private Variant serialize ()
  {
    return _raw.serialize ();
  }

  // preserving ABI compability in Vala? why not!
  internal virtual void dummy1 () {}
  internal virtual void dummy2 () {}
  internal virtual void dummy3 () {}
  internal virtual void dummy4 () {}
}

/* This is 1:1 copy of Protocol.LayoutHint, but we need to expose this
 * to our gir, we don't want to depend on UnityProtocol's gir */
public enum LayoutHint
{
  NONE,
  LEFT,
  RIGHT,
  TOP,
  BOTTOM
}

public class PreviewAction : Object, Dee.Serializable // TODO: Implement GLib.Action
{
  public string id { get; construct; }
  public string display_name { get; construct; }
  public Icon? icon_hint { get; construct; }
  public LayoutHint layout_hint { get; construct; }

  public PreviewAction (string id, string display_name, Icon? icon_hint)
  {
    Object (id: id, display_name: display_name, icon_hint: icon_hint);
  }

  public PreviewAction.with_layout_hint (string id, string display_name,
                                         Icon? icon_hint, LayoutHint layout)
  {
    Object (id: id, display_name: display_name, icon_hint: icon_hint,
            layout_hint: layout);
  }

  public signal ActivationResponse activated (string uri);

  private Variant serialize ()
  {
    // FIXME: we should use PreviewActionRaw, but this is faster
    Variant tuple[5];

    tuple[0] = id;
    tuple[1] = display_name;
    tuple[2] = new Variant.string (icon_hint != null ? icon_hint.to_string () : "");
    tuple[3] = (uint) layout_hint;
    var hints = new HashTable<string, Variant> (null, null);
    tuple[4] = hints;

    return new Variant.tuple (tuple);
  }

  static construct
  {
    Dee.Serializable.register_parser (typeof (PreviewAction),
                                      new VariantType ("(sssua{sv})"),
                                      (data) =>
    {
      unowned string icon_hint = data.get_child_value (2).get_string ();
      Icon? icon = null;
      if (icon_hint != null && icon_hint != "")
      {
        try
        {
          icon = Icon.new_for_string (icon_hint);
        }
        catch (Error err)
        {
          warning ("Failed to deserialize GIcon: %s", err.message);
        }
      }

      return new PreviewAction.with_layout_hint (
        data.get_child_value (0).get_string (),
        data.get_child_value (1).get_string (),
        icon,
        (LayoutHint) data.get_child_value (3).get_uint32 ());
    });
  }

  // preserving ABI compability in Vala? why not!
  internal virtual void dummy1 () {}
  internal virtual void dummy2 () {}
  internal virtual void dummy3 () {}
  internal virtual void dummy4 () {}
}


public class InfoHint : InitiallyUnowned
{
  public string id { get; construct; }
  public string display_name { get; construct; }
  public Icon? icon_hint { get; construct; }
  public Variant data { get; construct; }

  public InfoHint (string id, string display_name, Icon? icon_hint,
                   string data)
  {
    Object (id: id, display_name: display_name, icon_hint: icon_hint,
            data: new Variant.string (data));
  }

  public InfoHint.with_variant (string id, string display_name,
                                Icon? icon_hint, Variant data)
  {
    Object (id: id, display_name: display_name, icon_hint: icon_hint,
            data: data);
  }
}


public class GenericPreview : Preview
{
  public GenericPreview (string title,
                         string description,
                         Icon? image)
  {
    Object (title: title, image: image,
            description_markup: description);
  }

  internal override Object create_raw ()
  {
    return new Protocol.GenericPreview ();
  }
}


public class ApplicationPreview : Preview
{
  public Icon app_icon
  {
    get { return _raw.app_icon; }
    set { _raw.app_icon = value; }
  }
  public string license
  {
    get { return _raw.license; }
    set { _raw.license = value; }
  }
  public string copyright
  {
    get { return _raw.copyright; }
    set { _raw.copyright = value; }
  }
  public string last_update
  {
    get { return _raw.last_update; }
    set { _raw.last_update = value; }
  }

  public ApplicationPreview (string title,
                             string subtitle,
                             string description,
                             Icon? icon,
                             Icon? screenshot)
  {
    Object (title: title, subtitle: subtitle, image: screenshot,
            description_markup: description, app_icon: icon);
  }

  public void set_rating (float rating, uint num_ratings)
  {
    _raw.rating = rating;
    _raw.num_ratings = num_ratings;
  }

  private unowned Protocol.ApplicationPreview _raw;
  internal override Object create_raw ()
  {
    var raw = new Protocol.ApplicationPreview ();
    _raw = raw;
    return _raw;
  }
}


public class MusicPreview : Preview
{
  /* Keep in sync with Protocol.PlayState! */
  public enum TrackState
  {
    STOPPED,
    PLAYING,
    PAUSED
  }

  private Dee.SharedModel _track_data;
  private float _current_progress;
  private string _current_track_uri;
  private TrackState _current_track_state;

  public string current_track_uri
  {
    get { return _current_track_uri; }
    set
    {
      if (_current_track_uri != value)
      {
        _current_track_uri = value;
        update_track_state ();
      }
    }
  }
  public float current_progress
  {
    get { return _current_progress; }
    set { _current_progress = value; update_track_state (); }
  }
  public TrackState current_track_state
  {
    get { return _current_track_state; }
    set
    {
      _current_track_state = value;
      if (_current_track_state == TrackState.STOPPED)
      {
        _current_progress = 0.0f;
      }
      update_track_state ();
    }
  }

  public MusicPreview (string title,
                       string subtitle,
                       Icon? image)
  {
    Object (title: title, subtitle: subtitle, image: image);
  }

  private unowned Protocol.MusicPreview _raw;
  internal override Object create_raw ()
  {
    var raw = new Protocol.MusicPreview ();
    _raw = raw;
    return _raw;
  }

  public void add_track (TrackMetadata track)
  {
    init_model ();

    _track_data.append (track.uri, track.track_no, track.title,
                        track.length, TrackState.STOPPED, 0.0);
  }

  private void update_track_state () requires (_track_data != null)
  {
    var model = _track_data;
    var iter = model.get_first_iter ();
    var end_iter = model.get_last_iter ();

    bool found_last_playing = false;
    // linear search... kindof eek, but there should be only ~20 tracks
    while (iter != end_iter)
    {
      if (model.get_string (iter, TrackDataColumns.URI) == _current_track_uri)
      {
        var row = model.get_row (iter);
        row[TrackDataColumns.PLAY_STATE] = new Variant.uint32 (_current_track_state);
        row[TrackDataColumns.PROGRESS] = (double) _current_progress;

        model.set_row (iter, row);
        if (found_last_playing) break;
      }
      else if (model.get_uint32 (iter, TrackDataColumns.PLAY_STATE) != TrackState.STOPPED)
      {
        found_last_playing = true;
        var row = model.get_row (iter);
        row[TrackDataColumns.PLAY_STATE] = new Variant.uint32 (TrackState.STOPPED);
        row[TrackDataColumns.PROGRESS] = 0.0;

        model.set_row (iter, row);
      }
      iter = model.next (iter);
    }
  }

  internal override HashTable<string, Variant> update_property (HashTable<string, Variant> values)
  {
    Variant? action_v = values["action"];
    Variant? uri_v = values["uri"];
    if (action_v != null && uri_v != null)
    {
      switch (action_v.get_string ())
      {
        case "play":
          play (uri_v.get_string ());
          break;
        case "pause":
          pause (uri_v.get_string ());
          break;
        default:
          warning ("Unknown MusicPreview action: %s", action_v.get_string ());
          break;
      }
    }
    return base.update_property (values);
  }

  public signal void play (string uri);
  public signal void pause (string uri);

  // use add_info to add total number of tracks and "tags"

  private enum TrackDataColumns
  {
    URI,
    TRACK_NO,
    TITLE,
    LENGTH,
    PLAY_STATE,
    PROGRESS
  }

  private void init_model ()
  {
    if (_track_data == null)
    {
      string name = ("com.canonical.Unity.Scope.TrackData.T%" + int64.FORMAT)
        .printf (get_monotonic_time ());
      // FIXME: we'll probably want LEADER_WRITABLE access mode
      _track_data = new Dee.SharedModel (name);
      _track_data.set_schema ("s", "i", "s", "u", "u", "d");

      _raw.track_data_swarm_name = name;
    }
  }
}


public class MoviePreview : Preview
{
  public string year
  {
    get { return _raw.year; }
    set { _raw.year = value; }
  }

  public MoviePreview (string title, string subtitle, string description,
                       Icon? image)
  {
    Object (title: title, subtitle: subtitle, description_markup: description,
            image: image);
  }

  public void set_rating (float rating, uint num_ratings)
  {
    _raw.rating = rating;
    _raw.num_ratings = num_ratings;
  }

  private unowned Protocol.MoviePreview _raw;
  internal override Object create_raw ()
  {
    var raw = new Protocol.MoviePreview ();
    _raw = raw;
    return _raw;
  }
}


public class SeriesItem : Object
{
  public string uri { get; construct; }
  public string title { get; construct; }
  public Icon? icon_hint { get; construct; }

  public SeriesItem (string uri, string title, Icon? icon_hint)
  {
    Object (uri: uri, title: title, icon_hint: icon_hint);
  }
}

public class SeriesPreview : Preview, Dee.Serializable
{
  private string _current_item_uri;
  public string current_item_uri
  {
    get { return _current_item_uri; }
    construct { _current_item_uri = value; }
  }

  // FIXME: why do we even keep this?
  private int _selected_item = -1;

  public SeriesPreview (SeriesItem[] items, string selected_item_uri)
  {
    // careful current_item_uri will be set before items are
    Object (current_item_uri: selected_item_uri);
    foreach (unowned SeriesItem item in items)
    {
      add_item (item);
    }
  }

  private void add_item (SeriesItem item)
  {
    _raw.add_series_item (item.title, item.uri, item.icon_hint);
  }

  /* Emits request_item_preview and updates internal state */
  private void update_child_preview (string uri, int item_index = -1)
  {
    var preview = request_item_preview (uri);
    if (preview != null)
    {
      if (item_index < 0)
      {
        // try to auto-update _selected_item
        int i = 0;
        foreach (unowned Protocol.SeriesItemRaw item in _raw.get_items ())
        {
          if (item.uri == uri)
          {
            _selected_item = i;
            break;
          }
          i++;
        }
        if (_selected_item < 0) warning ("Unable to find URI in series items");
      }
      else _selected_item = item_index;

      _child_preview = preview;
      _raw.child_preview = preview.get_raw ();
      if (_selected_item >= 0 && _raw.selected_item != _selected_item)
        _raw.selected_item = _selected_item;
      _current_item_uri = uri;
    }
  }

  private Preview _child_preview;

  public signal Preview? request_item_preview (string uri);

  internal override HashTable<string, Variant> update_property (HashTable<string, Variant> values)
  {
    var response = base.update_property (values);
    int new_selected_item = _raw.selected_item;
    Preview? preview = null;
    if (new_selected_item != _selected_item &&
        new_selected_item < _raw.get_items ().length)
    {
      unowned Protocol.SeriesItemRaw item = _raw.get_items ()[new_selected_item];
      update_child_preview (item.uri, new_selected_item);
    }
    preview = _child_preview;
    response["preview"] = preview.serialize ();
    return response;
  }

  internal Preview? get_active_preview ()
  {
    return _child_preview;
  }

  private Variant serialize ()
  {
    // ensures that we emit selected_item_changed shortly after construction
    if (_selected_item < 0 && _raw.get_items ().length > 0)
    {
      update_child_preview (current_item_uri);
    }
    return _raw.serialize ();
  }

  private unowned Protocol.SeriesPreview _raw;
  internal override Object create_raw ()
  {
    var raw = new Protocol.SeriesPreview ();
    _raw = raw;
    return _raw;
  }
}


// Keeping this internal so far, we'll probably expose it later
internal class AsyncPreview : Preview
{
  public AsyncPreview ()
  {
    Object ();
  }

  public Cancellable cancellable { get; private set; }

  public signal void preview_ready (Preview preview);
  // FIXME: also add an emit method, so it's easy to use in C/Python

  internal override Object create_raw ()
  {
    return new Protocol.GenericPreview ();
  }
}

} /* namespace */
